//
// Created by dawsh on 1/17/24.
//
#pragma once

#include <string>
#include <iostream>
#include <JGL/Color4.h>
#include <JGL/Gradient.h>
#include <JGL/FontCache.h>
#include <J3ML/LinearAlgebra.h>
#include <J3ML/LinearAlgebra/Vector2.h>
#include <J3ML/LinearAlgebra/Vector3.h>
#include <J3ML/Geometry/Sphere.h>
#include <J3ML/Geometry/Capsule.h>
#include <J3ML/Geometry/TriangleMesh.h>

// OpenGL Wrapper for rendering 2D graphics primitives in both a 2D and 3D context
namespace JGL {

    using J3ML::LinearAlgebra::Vector2;
    using J3ML::LinearAlgebra::Vector3;
    using J3ML::LinearAlgebra::Matrix3x3;
    using J3ML::LinearAlgebra::Matrix4x4;
    using J3ML::LinearAlgebra::AxisAngle;
    using J3ML::LinearAlgebra::Quaternion;

    using J3ML::Geometry::Sphere;
    using J3ML::Geometry::OBB;
    using J3ML::Geometry::Capsule;
    using J3ML::Geometry::TriangleMesh;
    using J3ML::Geometry::Plane;

    struct HSV {
        float hue;
        float saturation;
        float value;
    };

    struct Triangle2D
    {
        Vector2 A;
        Vector2 B;
        Vector2 C;
    };

    struct Triangle3D
    {
        Vector3 A;
        Vector3 B;
        Vector3 C;
    };

    bool Update(const Vector2& window_size);
    bool InitTextEngine();
    int LoadFont(const std::string& font_path);
    void PurgeFontCache();

    void UnloadFont(int font_index);
    // TODO: implement correct coloring

    /// Drawing functions for primitive 2D Shapes.
    /// Each function is overloaded with Color3 and Color4 for optional transparency.
    namespace J2D {

        /// Open a 2-D rendering context with the underlying graphics system (In this case & by default OpenGL).
        /// @note This call may not strictly be necessary on some setups, but is provided to keep the API constant.
        /// It is recommended to always open a JGL 2D context to render your content, then close when completed.
        /// This keeps our code from, say, clobbering the OpenGL rendering context driving 3D content in between our calls.
        void Begin();
        /// Closes a 2-D rendering context with the underlying graphics system (In this case & by default OpenGL).
        /// @see Begin().
        void End();

        /// Plots a single pixel on the screen.
        /// @param color  A 3-or-4 channel color value. @see classes Color3, Color4
        /// @param coordinates The pixel-point on-screen at which to plot the pixel.
        void DrawPixel(const Color3& color, const Vector2& coordinates);
        void DrawPixel(const Color3& color, float x, float y);
        void DrawPixel(const Color4& color, const Vector2& coordinates);
        void DrawPixel(const Color4& color, float x, float y);

        /// Plots a line (segment) on the screen.
        /// @param color A 3-or-4 channel color value. @see classes Color3, Color4.
        /// @param A The starting point of the line segment.
        /// @param B The end point of the line segment.
        /// @param thickness The width at which to render the line.
        void DrawLine(const Color3& color, const Vector2& A, const Vector2& B, float thickness = 1);
        void DrawLine(const Color3 &color, float x, float y, float w, float h, float thickness = 1);
        void DrawLine(const Color4& color, const Vector2& A, const Vector2& B, float thickness = 1);
        void DrawLine(const Color4& color, float x1, float y1, float x2, float y2, float thickness = 1);

        ///Draws a line with a gradient that transitions across it.
        void DrawGradientLine(const Color4& color1, const Color4& color2, const Vector2& A, const Vector2& B, float thickness = 1);
        void DrawGradientLine(const Color3& color1, const Color3& color2, const Vector2& A, const Vector2& B, float thickness = 1);
        void DrawGradientLine(const Color4& color1, const Color4& color2, float x, float y, float w, float h, float thickness = 1);
        void DrawGradientLine(const Color3& color1, const Color3& color2, float x, float y, float w, float h, float thickness = 1);

        /// Draws an outline of a rectangle on the screen.
        void OutlineRect(const Color4& color, const Vector2& pos, const Vector2& size, float thickness = 1);
        void OutlineRect(const Color3& color, const Vector2& pos, const Vector2& size, float thickness = 1);

        /// Draws a filled rectangle on the screen.
        void FillRect(const Color4& color, const Vector2& pos, const Vector2& size);
        void FillRect(const Color3& color, const Vector2& pos, const Vector2& size);

        /// Draws a filled rectangle where the color transitions across it.
        void FillGradientRect(const Color4& color1, const Color4& color2, const Gradient& gradient, const Vector2& pos, const Vector2& size);
        void FillGradientRect(const Color3& color1, const Color3& color2, const Gradient& gradient, const Vector2& pos, const Vector2& size);

        /// Draws a filled rectangle with rounded corners on the screen.
        void FillRoundedRect(const Color4 &color, const Vector2 &pos, const Vector2 &size, float radius = 5, unsigned int subdivisions = 8);
        void FillRoundedRect(const Color3& color, const Vector2& pos, const Vector2& size, float radius = 5, unsigned int subdivisions = 8);

        /// Draws an outline of a circle on the screen.
        void OutlineCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions = 16, float thickness = 1);
        void OutlineCircle(const Color3& color, const Vector2& center, float radius, unsigned int subdivisions = 16, float thickness = 1);

        /// Draws a filled circle on the screen.
        void FillCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions = 8);
        void FillCircle(const Color3& color, const Vector2& center, float radius, unsigned int subdivisions = 8);

        /// Draws an outline of a triangle on the screen.
        void OutlineTriangle(const Color4& color, const Triangle2D& tri, float thickness = 1);
        void OutlineTriangle(const Color3& color, const Triangle2D& tri, float thickness = 1);
        // TODO: Implement an overload that simply takes 3 Vector3's

        /// Draws a filled triangle on the screen.
        void FillTriangle(const Color4& color, const Triangle2D &tri);
        void FillTriangle(const Color3& color, const Triangle2D& tri);
        // TODO: Implement an overload that simply takes 3 Vector3's

        /// Draws a text string on the screen with a given point-size and font.
        void DrawString(const Color4& color, const std::string& text, float x, float y, float scale, u32 size, unsigned int font_index);
        void DrawString(const Color3& color, const std::string& text, float x, float y, float scale, u32 size, unsigned int font_index);

        // TODO: Implement the following:
        void FillTexturedTriangle();
        void FillTexturedPolygon();
        void DrawSprite();
        void DrawPartialSprite();
        void DrawCubicBezierCurve();
        void OutlinePolygon    (const Color4& color, std::vector<Vector2> points);
        void FillPolygon       (const Color4& color, std::vector<Vector2> points, float thickness = 1);
        void OutlineRoundedRect(const Color4& color, const Vector2& pos, const Vector2& size, float radius = 5, float thickness = 1);
    }

    namespace J3D {
        void Begin();
        void End();
        void SetMatrix(const std::vector<GLfloat>& matrix, const Vector2& window_size);
        void DrawLine(const Color4& color, const Vector3& A, const Vector3& B, float thickness = 1);
        void DrawLine(const Color3& color, const Vector3& A, const Vector3& B, float thickness = 1);
        void FillSphere(const Color3& color, const Sphere& sphere);
        void WireframeSphere(const Color3& color, const Sphere& sphere, float thickness = 1);
        void FillOBB(const Color3& color, const OBB& obb);
        void WireframeOBB(const Color3& color, const OBB& obb, float thickness = 1);
        void FillCapsule(const Color3& color, const Capsule& capsule);
        void WireframeCapsule(const Color3& color, const Capsule& cap, float thickness = 1);
        void FillTriangleMesh(const Color3& color, const TriangleMesh& mesh);
        void WireframeTriangleMesh(const Color3& color, const TriangleMesh& mesh, float thickness = 1);
        void DrawString(const Color3& color, const std::string& text, const Vector3& pos, const Vector3& angle, float scale, u32 size, unsigned int font_index);

        void DrawMatrixGizmo     (const Matrix3x3&, const Vector3&);
        void DrawMatrixGizmo     (const Matrix4x4&);
        void DrawAxisAngleGizmo  (const AxisAngle&,  const Vector3&);
        void DrawQuaternionGizmo (const Quaternion&, const Vector3&);
    }
}