//
// Created by dawsh on 1/17/24.
//

#include <JGL/JGL.h>
#include <glad/glad.h>
#include <JGL/Color3.h>
#include <jlog/jlog.hpp>

GLfloat oldColor[4] = {0, 0, 0, 255};
bool inJ2D = false;
bool inJ3D = false;
bool wasTexture2DEnabled = false;
bool wasTextureCoordArrayEnabled = false;
bool wasDepthTestEnabled = false;
bool wasVertexArraysEnabled = false;
bool wasCullFaceEnabled = false;
bool wasBlendEnabled = false;
bool wasColorArrayEnabled = false;
GLint activeTextureUnit = 0;

namespace JGL {
    using namespace J3ML;
    Vector2 wS;

    bool Update(const Vector2& window_size) {
        wS = window_size;
        return JGL::InitTextEngine();
    }

    void J2D::Begin() {
        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity();
        glOrtho(0, wS.x, wS.y, 0, -1, 1);
        glMatrixMode(GL_MODELVIEW);
        glPushMatrix();
        glLoadIdentity();

        glGetIntegerv(GL_ACTIVE_TEXTURE, &activeTextureUnit);
        activeTextureUnit = activeTextureUnit - GL_TEXTURE0;
        if (activeTextureUnit != 0)
            glActiveTexture(GL_TEXTURE0);

        wasDepthTestEnabled = false;
        if (glIsEnabled(GL_DEPTH_TEST))
            wasDepthTestEnabled = true,
            glDisable(GL_DEPTH_TEST);

        wasVertexArraysEnabled = true;
        if (!glIsEnabled(GL_VERTEX_ARRAY))
            wasVertexArraysEnabled = false,
            glEnable(GL_VERTEX_ARRAY);

        wasCullFaceEnabled = true;
        if (!glIsEnabled(GL_CULL_FACE))
            wasCullFaceEnabled = false,
            glEnable(GL_CULL_FACE),
            glCullFace(GL_BACK);

        wasBlendEnabled = true;
        if (!glIsEnabled(GL_BLEND))
            wasBlendEnabled = false,
            glEnable(GL_BLEND),
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        wasTexture2DEnabled = true;
        if (!glIsEnabled(GL_TEXTURE_2D))
            wasTexture2DEnabled = false,
            glEnable(GL_TEXTURE_2D);

        wasTextureCoordArrayEnabled = true;
        if (!glIsEnabled(GL_TEXTURE_COORD_ARRAY))
            wasTextureCoordArrayEnabled = false,
            glEnable(GL_TEXTURE_COORD_ARRAY);

        wasColorArrayEnabled = false;
        if (glIsEnabled(GL_COLOR_ARRAY))
            wasColorArrayEnabled = true,
            glDisable(GL_COLOR_ARRAY);

        if (!inJ3D)
            inJ2D = true;
        else { ERROR("Attempt to Begin J2D inside of J3D context.") }
    }

    void J2D::End() {
        //Change back to the previous projection (The 3D one in Re3D's case.)
        glDisable(GL_TEXTURE_2D);
        glPopMatrix();
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
        if (wasDepthTestEnabled)
            glEnable(GL_DEPTH_TEST);

        if (!wasVertexArraysEnabled)
            glDisable(GL_VERTEX_ARRAY);

        if (!wasCullFaceEnabled)
            glDisable(GL_CULL_FACE);

        if (!wasBlendEnabled)
            glDisable(GL_BLEND);

        if (!wasTexture2DEnabled)
            glDisable(GL_TEXTURE_2D);

        if (!wasTextureCoordArrayEnabled)
            glDisable(GL_TEXTURE_COORD_ARRAY);

        if (wasColorArrayEnabled)
            glEnable(GL_COLOR_ARRAY);

        //Select whatever texture unit was selected before.
        glActiveTexture(GL_TEXTURE0 + activeTextureUnit);

        //Put the draw color back how it was before.
        glColor4f(oldColor[0], oldColor[1], oldColor[2], oldColor[3]);

        inJ2D = false;
    }

    void J2D::FillRect(const Color4& color, const Vector2& pos, const Vector2& size) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};
        glColor4f(color.r / 255.f, color.g / 255.f, color.b / 255.f, color.a / 255.f);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_QUADS, 0, 4);
    }

    void J2D::FillRect(const Color3& color, const Vector2& pos, const Vector2& size) {
        J2D::FillRect({color.r, color.g, color.b, 255}, pos, size);
    }

    void J2D::FillGradientRect(const Color4& color1, const Color4& color2, const Gradient& gradient, const Vector2& pos, const Vector2& size) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};
        std::vector<GLfloat> colors = {};

        if (gradient == Gradient::Horizontal)
            colors = {color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f, color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f,
                     color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f, color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f};

        else if (gradient == Gradient::Vertical)
            colors = {color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f, color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f,
                     color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f, color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f};

        else if (gradient == Gradient::DiagonalBottomLeft)
            colors = {(color1.r + color2.r) / 2.f / 255.f, (color1.g + color2.g) / 2.f / 255.f, (color1.b + color2.b) / 2.f / 255.f, (color1.a + color2.a) / 2.f / 255.f,
                    color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f,(color1.r + color2.r) / 2.f / 255.f, (color1.g + color2.g) / 2.f / 255.f,
                    (color1.b + color2.b) / 2.f / 255.f, (color1.a + color2.a) / 2.f / 255.f, color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f};

        else if (gradient == Gradient::DiagonalTopLeft)
                    colors = {color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f,(color1.r + color2.r) / 2.f / 255.f, (color1.g + color2.g) / 2.f / 255.f,
                              (color1.b + color2.b) / 2.f / 255.f, (color1.a + color2.a) / 2.f / 255.f,color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f,
                              (color1.r + color2.r) / 2.f / 255.f, (color1.g + color2.g) / 2.f / 255.f, (color1.b + color2.b) / 2.f / 255.f,(color1.a + color2.a) / 2.f / 255.f};

        glEnable(GL_COLOR_ARRAY);
            glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
            glColorPointer(4, GL_FLOAT, 0, colors.data());
            glDrawArrays(GL_QUADS, 0, 4);
        glDisable(GL_COLOR_ARRAY);
    }

    void J2D::FillGradientRect(const Color3& color1, const Color3& color2, const Gradient& gradient, const Vector2& pos, const Vector2& size) {
        J2D::FillGradientRect({color1.r, color1.g, color1.b, 255}, {color2.r, color2.g, color2.b, 255}, gradient, pos, size);
    }

    void J2D::FillRoundedRect(const Color4 &color, const Vector2 &pos, const Vector2 &size, float radius, unsigned int subdivisions) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        J2D::FillRect(color, {pos.x + radius, pos.y}, {size.x - 2 * radius, size.y});
        J2D::FillRect(color, {pos.x, pos.y + radius}, {size.x, size.y - 2 * radius});

        J2D::FillCircle(color, {pos.x + radius, pos.y + radius}, radius, subdivisions);
        J2D::FillCircle(color, {pos.x + size.x - radius, pos.y + radius}, radius, subdivisions);
        J2D::FillCircle(color, {pos.x + radius, pos.y + size.y - radius}, radius, subdivisions);
        J2D::FillCircle(color, {pos.x + size.x - radius, pos.y + size.y - radius}, radius, subdivisions);
    }

    void J2D::FillRoundedRect(const JGL::Color3& color, const J3ML::LinearAlgebra::Vector2& pos, const J3ML::LinearAlgebra::Vector2& size, float radius, unsigned int subdivisions) {
        J2D::FillRoundedRect({color.r, color.g, color.b, 255}, pos, size, radius, subdivisions);
    }

    void J2D::OutlineRect(const Color4& color, const Vector2& pos, const Vector2& size, float thickness) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};

        glLineWidth(thickness);
        glColor4f(color.r / 255.f, color.g / 255.f, color.b / 255.f, color.a / 255.f);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINE_LOOP, 0, 4);
    }

    void J2D::OutlineRect(const Color3& color, const Vector2& pos, const Vector2& size, float thickness) {
        J2D::OutlineRect({color.r, color.g, color.b, 255}, pos, size, thickness);
    }

    void J2D::DrawLine(const Color4& color, const Vector2& A, const Vector2& B, float thickness) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.");

        Vector2 vertices[] = {A, B};

        glLineWidth(thickness);
        glColor4f(color.r / 255.f, color.g / 255.f, color.b / 255.f, color.a / 255.f);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINES, 0, 2);
    }

    void J2D::DrawLine(const Color3& color, const Vector2& A, const Vector2& B, float thickness) {
        J2D::DrawLine({color.r, color.g, color.b, 255}, A, B, thickness);
    }

    void J2D::DrawLine(const Color4& color, float x, float y, float w, float h, float thickness) {
        J2D::DrawLine(color, {x, y}, {w, h}, thickness);
    }

    void J2D::DrawLine(const Color3& color, float x, float y, float w, float h, float thickness) {
        J2D::DrawLine({color.r, color.g, color.b, 255}, x, y, w, h, thickness);
    }

    void J2D::DrawGradientLine(const Color4& color1, const Color4& color2, const Vector2& A, const Vector2& B, float thickness) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.");

        Vector2 vertices[] = {A, B};
        GLfloat colors[8] = {color1.r / 255.f, color1.g / 255.f, color1.b / 255.f, color1.a / 255.f,
                             color2.r / 255.f, color2.g / 255.f, color2.b / 255.f, color2.a / 255.f};

        glEnable(GL_COLOR_ARRAY);
            glLineWidth(thickness);
            glColorPointer(4,GL_FLOAT,sizeof(GL_FLOAT) * 4, colors);
            glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
            glDrawArrays(GL_LINES, 0, 2);
        glDisable(GL_COLOR_ARRAY);
    }

    void J2D::DrawGradientLine(const Color3& color1, const Color3& color2, const Vector2& A, const Vector2& B, float thickness) {
        J2D::DrawGradientLine({color1.r, color1.g, color1.b, 255}, {color2.r, color2.g, color2.b, 255}, A, B, thickness);
    }
    void DrawGradientLine(const Color4& color1, const Color4& color2, float x, float y, float w, float h, float thickness) {
        J2D::DrawGradientLine(color1, color2, {x, y}, {w, h}, thickness);
    }

    void DrawGradientLine(const Color3& color1, const Color3& color2, float x, float y, float w, float h, float thickness) {
        J2D::DrawGradientLine({color1.r, color1.g, color1.b, 255}, {color2.r, color2.g, color2.b, 255}, {x, y}, {w, h}, thickness);
    }

    void J2D::DrawPixel(const Color4& color, const Vector2& coordinates) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.");

        Vector2 vertices[] = {coordinates};

        glColor4f(color.r / 255.f, color.g / 255.f, color.b / 255.f, color.a / 255.f);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINES, 0, 1);
    }

    void J2D::DrawPixel(const Color3& color, const Vector2& coordinates) {
        J2D::DrawPixel({color.r, color.g, color.b, 255}, coordinates);
    }

    void J2D::DrawPixel(const Color4& color, float x, float y) {
        DrawPixel(color, {x, y});
    }

    void J2D::DrawPixel(const Color3& color, float x, float y) {
        DrawPixel({color.r, color.g, color.b, 255}, {x, y});
    }

    void J2D::OutlineCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions, float thickness) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        float step = (2.f * M_PI) / (float) subdivisions;
        std::vector<Vector2> vertices{};
        GLfloat angle, x, y;

        for (angle = 0.0f; angle < (2.f * M_PI); angle += step) {
            x = radius * std::sin(angle) + center.x;
            y = radius * std::cos(angle) + center.y;
            vertices.emplace_back(x,y);
        }

        glLineWidth(thickness);
        glColor4f(color.r / 255.f, color.g / 255.f, color.b / 255.f, color.a / 255.f);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glDrawArrays(GL_LINE_LOOP, 0, vertices.size());
    }

    void J2D::OutlineCircle(const Color3& color, const Vector2& center, float radius, unsigned int subdivisions, float thickness) {
        J2D::OutlineCircle({color.r, color.g, color.b, 255}, center, radius, subdivisions, thickness);
    }

    void J2D::FillCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        float step =  (2.f * M_PI) / (float) subdivisions;;
        std::vector<Vector2> vertices{};
        GLfloat angle, x, y;

        for (angle = 0.0f; angle < (2.f * M_PI); angle += step)
                x = radius * sin(angle) + center.x,
                y = radius * cos(angle) + center.y,
                vertices.push_back({x, y});

        glColor4f(color.r / 255.f, color.g / 255.f, color.b / 255.f, color.a / 255.f);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
        glDrawArrays(GL_TRIANGLE_FAN, 0, vertices.size());
    }

    void J2D::FillCircle(const Color3& color, const Vector2& center, float radius, unsigned int subdivisions) {
        J2D::FillCircle({color.r, color.g, color.b, 255}, center, radius, subdivisions);
    }

    void J2D::OutlineTriangle(const Color4& color, const Triangle2D& tri, float thickness) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};

        glLineWidth(thickness);
        glColor4f(color.r / 255.f, color.g / 255.f, color.b / 255.f, color.a / 255.f);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_LINE_LOOP, 0, 3);
    }

    void J2D::OutlineTriangle(const Color3& color, const Triangle2D& tri, float thickness) {
        J2D::OutlineTriangle({color.r, color.g, color.b, 255}, tri, thickness);
    }

    void J2D::FillTriangle(const Color4& color, const Triangle2D &tri) {
        if (!inJ2D)
            ERROR("Attempt to Render J2D element before J2D begin.")

        Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};

        glColor4f(color.r / 255.f, color.g / 255.f, color.b / 255.f, color.a / 255.f);
        glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
        glDrawArrays(GL_TRIANGLES, 0, 3);
    }

    void J2D::FillTriangle(const Color3& color, const Triangle2D &tri) {
        J2D::FillTriangle({color.r, color.g, color.b, 255}, tri);
    }

    void J3D::Begin() {

        //Get what the draw color was before we did anything.
        glGetFloatv(GL_CURRENT_COLOR, oldColor);

        wasDepthTestEnabled = false;
        if (glIsEnabled(GL_DEPTH_TEST))
            wasDepthTestEnabled = true,
            glDisable(GL_DEPTH_TEST);

        wasVertexArraysEnabled = false;
        if (!glIsEnabled(GL_VERTEX_ARRAY))
            wasVertexArraysEnabled = false,
            glEnable(GL_VERTEX_ARRAY);

        wasTexture2DEnabled = true;
        if (!glIsEnabled(GL_TEXTURE_2D))
            wasTexture2DEnabled = false,
            glEnable(GL_TEXTURE_2D);

        //TODO We won't always want this but for now we do.
        wasCullFaceEnabled = false;
        if (glIsEnabled(GL_CULL_FACE))
            wasCullFaceEnabled = true,
            glDisable(GL_CULL_FACE);

        wasBlendEnabled = true;
        if (!glIsEnabled(GL_BLEND))
            wasBlendEnabled = false,
            glEnable(GL_BLEND),
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);



        if (!inJ2D)
            inJ3D = true;
        else { ERROR("Attempt to Begin J3D inside of J2D context.") }
    }

    void J3D::End() {
        glDisable(GL_TEXTURE_2D);

        if (wasDepthTestEnabled)
            glEnable(GL_DEPTH_TEST);

        if (!wasVertexArraysEnabled)
            glDisable(GL_VERTEX_ARRAY);

        if (wasTexture2DEnabled)
            glDisable(GL_TEXTURE_2D);

        if (wasBlendEnabled)
            glDisable(GL_BLEND);

        if (wasCullFaceEnabled)
            glEnable(GL_CULL_FACE);

        //Put the draw color back how it was before.
        glColor4f(oldColor[0], oldColor[1], oldColor[2], oldColor[3]);

        inJ3D = false;
    }

    void J3D::DrawLine(const Color4& color, const Vector3& A, const Vector3& B, float thickness) {
        if (!inJ3D)
            ERROR("Attempt to Render J3D element before J3D begin.")

        Vector3 vertices[] = {A, B};

        glLineWidth(thickness);
        glColor4f(color.r / 255.f, color.g / 255.f, color.b / 255.f, color.a / 255.f);
        glVertexPointer(3, GL_FLOAT, sizeof(Vector3), vertices);
        glDrawArrays(GL_LINES, 0, 2);
    }

    void J3D::DrawLine(const Color3& color, const Vector3& A, const Vector3& B, float thickness) {
        J3D::DrawLine({color.r, color.g, color.b, 255}, A, B, thickness);
    }
}
